//  _    ___      __              __________      _     __        ______________   __
// | |  / (_)____/ /___  ______ _/ / ____/ /_  __(_)___/ /____   /  ___/ __  / /  / /
// | | / / / ___/ __/ / / / __ `/ / /_  / / / / / / __  / ___/  / /___/ /_/ / /  / /
// | |/ / / /  / /_/ /_/ / /_/ / / __/ / / /_/ / / /_/ (__  )  / /_) / ____/ /__/ /
// |___/_/_/   \__/\__,_/\__,_/_/_/   /_/\__,_/_/\__,_/____/   \____/_/    \_____/
//
//////////////////////////////////////////////////////////////////////////
#ifndef D3Q27_KERNELS_H
#define D3Q27_KERNELS_H

//random numbers
#include <curand.h>
#include <curand_kernel.h>
#include <cuda.h>
#include <cuda_runtime.h>

#include "LBM/LB.h"


__global__ void LBCalcMac27( real* vxD,
                                        real* vyD,
                                        real* vzD,
                                        real* rhoD,
                                        unsigned int* neighborX,
                                        unsigned int* neighborY,
                                        unsigned int* neighborZ,
                                        unsigned int* geoD,
                                        unsigned long long numberOfLBnodes,
                                        real* DD,
                                        bool isEvenTimestep);

__global__ void LBCalcMacSP27( real* vxD,
                                          real* vyD,
                                          real* vzD,
                                          real* rhoD,
                                          real* pressD,
                                          unsigned int* geoD,
                                          unsigned int* neighborX,
                                          unsigned int* neighborY,
                                          unsigned int* neighborZ,
                                          unsigned long long numberOfLBnodes,
                                          real* DD,
                                          bool isEvenTimestep);

__global__ void LBCalcMacCompSP27( real* vxD,
                                              real* vyD,
                                              real* vzD,
                                              real* rhoD,
                                              real* pressD,
                                              unsigned int* geoD,
                                              unsigned int* neighborX,
                                              unsigned int* neighborY,
                                              unsigned int* neighborZ,
                                              unsigned long long numberOfLBnodes,
                                              real* DD,
                                              bool isEvenTimestep);

__global__ void CalcConc7( real* Conc,
                                          unsigned int* geoD,
                                          unsigned int* neighborX,
                                          unsigned int* neighborY,
                                          unsigned int* neighborZ,
                                          unsigned long long numberOfLBnodes,
                                          real* DD7,
                                          bool isEvenTimestep);

__global__ void GetPlaneConc7(real* Conc,
                                            int* kPC,
                                            unsigned int numberOfPointskPC,
                                            unsigned int* geoD,
                                            unsigned int* neighborX,
                                            unsigned int* neighborY,
                                            unsigned int* neighborZ,
                                            unsigned long long numberOfLBnodes,
                                            real* DD7,
                                            bool isEvenTimestep);

__global__ void GetPlaneConc27(real* Conc,
                                             int* kPC,
                                             unsigned int numberOfPointskPC,
                                             unsigned int* geoD,
                                             unsigned int* neighborX,
                                             unsigned int* neighborY,
                                             unsigned int* neighborZ,
                                             unsigned long long numberOfLBnodes,
                                             real* DD27,
                                             bool isEvenTimestep);

__global__ void CalcConc27(real* Conc,
                                          unsigned int* geoD,
                                          unsigned int* neighborX,
                                          unsigned int* neighborY,
                                          unsigned int* neighborZ,
                                          unsigned long long numberOfLBnodes,
                                          real* DD27,
                                          bool isEvenTimestep);

__global__ void LBCalcMedSP27( real* vxD,
                                          real* vyD,
                                          real* vzD,
                                          real* rhoD,
                                          real* pressD,
                                          unsigned int* geoD,
                                          unsigned int* neighborX,
                                          unsigned int* neighborY,
                                          unsigned int* neighborZ,
                                          unsigned long long numberOfLBnodes,
                                          real* DD,
                                          bool isEvenTimestep);

__global__ void LBCalcMedCompSP27( real* vxD,
                                              real* vyD,
                                              real* vzD,
                                              real* rhoD,
                                              real* pressD,
                                              unsigned int* geoD,
                                              unsigned int* neighborX,
                                              unsigned int* neighborY,
                                              unsigned int* neighborZ,
                                              unsigned long long numberOfLBnodes,
                                              real* DD,
                                              bool isEvenTimestep);

__global__ void LBCalcMedCompAD27(
    real* vxD,
    real* vyD,
    real* vzD,
    real* rhoD,
    real* pressD,
    real* concD,
    unsigned int* geoD,
    unsigned int* neighborX,
    unsigned int* neighborY,
    unsigned int* neighborZ,
    unsigned long long numberOfLBnodes,
    real* DD,
    real* DD_AD,
    bool isEvenTimestep);

__global__ void LBCalcMacMedSP27( real* vxD,
                                             real* vyD,
                                             real* vzD,
                                             real* rhoD,
                                             real* pressD,
                                             unsigned int* geoD,
                                             unsigned int* neighborX,
                                             unsigned int* neighborY,
                                             unsigned int* neighborZ,
                                             unsigned int tdiff,
                                             unsigned long long numberOfLBnodes,
                                             bool isEvenTimestep);

__global__ void LBResetMedianValuesSP27(
    real* vxD,
    real* vyD,
    real* vzD,
    real* rhoD,
    real* pressD,
    unsigned long long numberOfLBnodes,
    bool isEvenTimestep);

__global__ void LBResetMedianValuesAD27(
    real* vxD,
    real* vyD,
    real* vzD,
    real* rhoD,
    real* pressD,
    real* concD,
    unsigned long long numberOfLBnodes,
    bool isEvenTimestep);

__global__ void LBCalc2ndMomentsIncompSP27(  real* kxyFromfcNEQ,
                                                        real* kyzFromfcNEQ,
                                                        real* kxzFromfcNEQ,
                                                        real* kxxMyyFromfcNEQ,
                                                        real* kxxMzzFromfcNEQ,
                                                        unsigned int* geoD,
                                                        unsigned int* neighborX,
                                                        unsigned int* neighborY,
                                                        unsigned int* neighborZ,
                                                        unsigned long long numberOfLBnodes,
                                                        real* DD,
                                                        bool isEvenTimestep);

__global__ void LBCalc2ndMomentsCompSP27(real* kxyFromfcNEQ,
                                                    real* kyzFromfcNEQ,
                                                    real* kxzFromfcNEQ,
                                                    real* kxxMyyFromfcNEQ,
                                                    real* kxxMzzFromfcNEQ,
                                                    unsigned int* geoD,
                                                    unsigned int* neighborX,
                                                    unsigned int* neighborY,
                                                    unsigned int* neighborZ,
                                                    unsigned long long numberOfLBnodes,
                                                    real* DD,
                                                    bool isEvenTimestep);

__global__ void LBCalc3rdMomentsIncompSP27(  real* CUMbbb,
                                                        real* CUMabc,
                                                        real* CUMbac,
                                                        real* CUMbca,
                                                        real* CUMcba,
                                                        real* CUMacb,
                                                        real* CUMcab,
                                                        unsigned int* bcMatD,
                                                        unsigned int* neighborX,
                                                        unsigned int* neighborY,
                                                        unsigned int* neighborZ,
                                                        real* DDStart,
                                                        unsigned long long numberOfLBnodes,
                                                        bool EvenOrOdd);

__global__ void LBCalc3rdMomentsCompSP27(real* CUMbbb,
                                                    real* CUMabc,
                                                    real* CUMbac,
                                                    real* CUMbca,
                                                    real* CUMcba,
                                                    real* CUMacb,
                                                    real* CUMcab,
                                                    unsigned int* bcMatD,
                                                    unsigned int* neighborX,
                                                    unsigned int* neighborY,
                                                    unsigned int* neighborZ,
                                                    real* DDStart,
                                                    unsigned long long numberOfLBnodes,
                                                    bool EvenOrOdd);

__global__ void LBCalcHigherMomentsIncompSP27(   real* CUMcbb,
                                                            real* CUMbcb,
                                                            real* CUMbbc,
                                                            real* CUMcca,
                                                            real* CUMcac,
                                                            real* CUMacc,
                                                            real* CUMbcc,
                                                            real* CUMcbc,
                                                            real* CUMccb,
                                                            real* CUMccc,
                                                            unsigned int* bcMatD,
                                                            unsigned int* neighborX,
                                                            unsigned int* neighborY,
                                                            unsigned int* neighborZ,
                                                            real* DDStart,
                                                            unsigned long long numberOfLBnodes,
                                                            bool EvenOrOdd);

__global__ void LBCalcHigherMomentsCompSP27( real* CUMcbb,
                                                        real* CUMbcb,
                                                        real* CUMbbc,
                                                        real* CUMcca,
                                                        real* CUMcac,
                                                        real* CUMacc,
                                                        real* CUMbcc,
                                                        real* CUMcbc,
                                                        real* CUMccb,
                                                        real* CUMccc,
                                                        unsigned int* bcMatD,
                                                        unsigned int* neighborX,
                                                        unsigned int* neighborY,
                                                        unsigned int* neighborZ,
                                                        real* DDStart,
                                                        unsigned long long numberOfLBnodes,
                                                        bool EvenOrOdd);

__global__ void LBCalcMeasurePoints(real* vxMP,
                                               real* vyMP,
                                               real* vzMP,
                                               real* rhoMP,
                                               unsigned int* kMP,
                                               unsigned int numberOfPointskMP,
                                               unsigned int MPClockCycle,
                                               unsigned int t,
                                               unsigned int* geoD,
                                               unsigned int* neighborX,
                                               unsigned int* neighborY,
                                               unsigned int* neighborZ,
                                               unsigned long long numberOfLBnodes,
                                               real* DD,
                                               bool isEvenTimestep);

__global__ void LB_BC_Press_East27( int nx,
                                               int ny,
                                               int tz,
                                               unsigned int* bcMatD,
                                               unsigned int* neighborX,
                                               unsigned int* neighborY,
                                               unsigned int* neighborZ,
                                               real* DD,
                                               unsigned long long numberOfLBnodes,
                                               bool isEvenTimestep) ;

__global__ void LB_BC_Vel_West_27( int nx,
                                              int ny,
                                              int nz,
                                              int itz,
                                              unsigned int* bcMatD,
                                              unsigned int* neighborX,
                                              unsigned int* neighborY,
                                              unsigned int* neighborZ,
                                              real* DD,
                                              unsigned long long numberOfLBnodes,
                                              bool isEvenTimestep,
                                              real u0x,
                                              unsigned int grid_nx,
                                              unsigned int grid_ny,
                                              real om);

//no Slip BCs
__global__ void QDevice27(real* distributions,
                                     int* subgridDistanceIndices,
                                     real* subgridDistances,
                                     unsigned int numberOfBCnodes,
                                     real omega,
                                     unsigned int* neighborX,
                                     unsigned int* neighborY,
                                     unsigned int* neighborZ,
                                     unsigned long long numberOfLBnodes,
                                     bool isEvenTimestep);

__global__ void QDeviceComp27(
                                         real* distributions,
                                         int* subgridDistanceIndices,
                                         real* subgridDistances,
                                         unsigned int numberOfBCnodes,
                                         real omega,
                                         unsigned int* neighborX,
                                         unsigned int* neighborY,
                                         unsigned int* neighborZ,
                                         unsigned long long numberOfLBnodes,
                                         bool isEvenTimestep);

__global__ void QDeviceCompThinWallsPartOne27(real* DD,
                                                         int* k_Q,
                                                         real* QQ,
                                                         unsigned int numberOfBCnodes,
                                                         real om1,
                                                         unsigned int* neighborX,
                                                         unsigned int* neighborY,
                                                         unsigned int* neighborZ,
                                                         unsigned long long numberOfLBnodes,
                                                         bool isEvenTimestep);

__global__ void QDevice3rdMomentsComp27(     real* distributions,
                                                     int* subgridDistanceIndices,
                                                     real* subgridDistances,
                                                     unsigned int numberOfBCnodes,
                                                     real omega,
                                                     unsigned int* neighborX,
                                                     unsigned int* neighborY,
                                                     unsigned int* neighborZ,
                                                     unsigned long long numberOfLBnodes,
                                                     bool isEvenTimestep);

__global__ void QDeviceIncompHighNu27(real* DD,
                                                 int* k_Q,
                                                 real* QQ,
                                                 unsigned int numberOfBCnodes,
                                                 real om1,
                                                 unsigned int* neighborX,
                                                 unsigned int* neighborY,
                                                 unsigned int* neighborZ,
                                                 unsigned long long numberOfLBnodes,
                                                 bool isEvenTimestep);

__global__ void QDeviceCompHighNu27(     real* DD,
                                                 int* k_Q,
                                                 real* QQ,
                                                 unsigned int numberOfBCnodes,
                                                 real om1,
                                                 unsigned int* neighborX,
                                                 unsigned int* neighborY,
                                                 unsigned int* neighborZ,
                                                 unsigned long long numberOfLBnodes,
                                                 bool isEvenTimestep);

//Velocity BCs
__global__ void QVelDevPlainBB27(
    real* velocityX,
    real* velocityY,
    real* velocityZ,
    real* distributions,
    int* subgridDistanceIndices,
    real* subgridDistances,
    uint numberOfBCnodes,
    uint* neighborX,
    uint* neighborY,
    uint* neighborZ,
    unsigned long long numberOfLBnodes,
    bool isEvenTimestep);

__global__ void QVelDevCouette27(real* vx,
                                            real* vy,
                                            real* vz,
                                            real* DD,
                                            int* k_Q,
                                            real* QQ,
                                            unsigned int numberOfBCnodes,
                                            real om1,
                                            unsigned int* neighborX,
                                            unsigned int* neighborY,
                                            unsigned int* neighborZ,
                                            unsigned long long numberOfLBnodes,
                                            bool isEvenTimestep);

__global__ void QVelDev1h27( int inx,
                                        int iny,
                                        real* vx,
                                        real* vy,
                                        real* vz,
                                        real* DD,
                                        int* k_Q,
                                        real* QQ,
                                        unsigned int numberOfBCnodes,
                                        real om1,
                                        real Phi,
                                        real angularVelocity,
                                        unsigned int* neighborX,
                                        unsigned int* neighborY,
                                        unsigned int* neighborZ,
                                        real* coordX,
                                        real* coordY,
                                        real* coordZ,
                                        unsigned long long numberOfLBnodes,
                                        bool isEvenTimestep);

__global__ void QVelDevice27(int inx,
                                        int iny,
                                        real* vx,
                                        real* vy,
                                        real* vz,
                                        real* DD,
                                        int* k_Q,
                                        real* QQ,
                                        unsigned int numberOfBCnodes,
                                        real om1,
                                        unsigned int* neighborX,
                                        unsigned int* neighborY,
                                        unsigned int* neighborZ,
                                        unsigned long long numberOfLBnodes,
                                        bool isEvenTimestep);

__global__ void QVelDeviceCompPlusSlip27(real* vx,
                                                    real* vy,
                                                    real* vz,
                                                    real* DD,
                                                    int* k_Q,
                                                    real* QQ,
                                                    unsigned int numberOfBCnodes,
                                                    real om1,
                                                    unsigned int* neighborX,
                                                    unsigned int* neighborY,
                                                    unsigned int* neighborZ,
                                                    unsigned long long numberOfLBnodes,
                                                    bool isEvenTimestep);

__global__ void QVelDeviceComp27(real* velocityX,
                                            real* velocityY,
                                            real* velocityZ,
                                            real* distribution,
                                            int* subgridDistanceIndices,
                                            real* subgridDistances,
                                            unsigned int numberOfBCnodes,
                                            real omega,
                                            unsigned int* neighborX,
                                            unsigned int* neighborY,
                                            unsigned int* neighborZ,
                                            unsigned long long numberOfLBnodes,
                                            bool isEvenTimestep);

__global__ void QVelDeviceCompThinWallsPartOne27(
    real* vx,
    real* vy,
    real* vz,
    real* DD,
    int* k_Q,
    real* QQ,
    uint numberOfBCnodes,
    real om1,
    uint* neighborX,
    uint* neighborY,
    uint* neighborZ,
    unsigned long long numberOfLBnodes,
    bool isEvenTimestep);

__global__ void QThinWallsPartTwo27(
    real* DD,
    int* k_Q,
    real* QQ,
    uint numberOfBCnodes,
    uint* geom,
    uint* neighborX,
    uint* neighborY,
    uint* neighborZ,
    uint* neighborWSB,
    unsigned long long numberOfLBnodes,
    bool isEvenTimestep);

__global__ void QVelDeviceCompZeroPress27(
    real* velocityX,
    real* velocityY,
    real* velocityZ,
    real* distribution,
    int* subgridDistanceIndices,
    real* subgridDistances,
    unsigned int numberOfBCnodes,
    real omega,
    unsigned int* neighborX,
    unsigned int* neighborY,
    unsigned int* neighborZ,
    unsigned long long numberOfLBnodes,
    bool isEvenTimestep);

__global__ void QVelDeviceIncompHighNu27(real* vx,
                                                    real* vy,
                                                    real* vz,
                                                    real* DD,
                                                    int* k_Q,
                                                    real* QQ,
                                                    unsigned int numberOfBCnodes,
                                                    real om1,
                                                    unsigned int* neighborX,
                                                    unsigned int* neighborY,
                                                    unsigned int* neighborZ,
                                                    unsigned long long numberOfLBnodes,
                                                    bool isEvenTimestep);

__global__ void QVelDeviceCompHighNu27(    real* vx,
                                                    real* vy,
                                                    real* vz,
                                                    real* DD,
                                                    int* k_Q,
                                                    real* QQ,
                                                    unsigned int numberOfBCnodes,
                                                    real om1,
                                                    unsigned int* neighborX,
                                                    unsigned int* neighborY,
                                                    unsigned int* neighborZ,
                                                    unsigned long long numberOfLBnodes,
                                                    bool isEvenTimestep);

__global__ void QVeloDeviceEQ27(real* VeloX,
                                           real* VeloY,
                                           real* VeloZ,
                                           real* DD,
                                           int* k_Q,
                                           int numberOfBCnodes,
                                           real om1,
                                           unsigned int* neighborX,
                                           unsigned int* neighborY,
                                           unsigned int* neighborZ,
                                           unsigned long long numberOfLBnodes,
                                           bool isEvenTimestep);

//Slip BCs
__global__ void QSlipDevice27(real* DD,
                                         int* k_Q,
                                         real* QQ,
                                         unsigned int numberOfBCnodes,
                                         real om1,
                                         unsigned int* neighborX,
                                         unsigned int* neighborY,
                                         unsigned int* neighborZ,
                                         unsigned long long numberOfLBnodes,
                                         bool isEvenTimestep);

__global__ void QSlipDeviceComp27(real* DD,
                                             int* k_Q,
                                             real* QQ,
                                             unsigned int numberOfBCnodes,
                                             real om1,
                                             unsigned int* neighborX,
                                             unsigned int* neighborY,
                                             unsigned int* neighborZ,
                                             unsigned long long numberOfLBnodes,
                                             bool isEvenTimestep);

__global__ void QSlipDeviceComp27TurbViscosity(
                                    real* distributions,
                                    int* subgridDistanceIndices,
                                    real* subgridDistances,
                                    unsigned int numberOfBCnodes,
                                    real omega,
                                    unsigned int* neighborX,
                                    unsigned int* neighborY,
                                    unsigned int* neighborZ,
                                    real* turbViscosity,
                                    unsigned long long numberOfLBnodes,
                                    bool isEvenTimestep);

__global__ void QSlipPressureDeviceComp27TurbViscosity(
                                    real* distributions,
                                    int* subgridDistanceIndices,
                                    real* subgridDistances,
                                    unsigned int numberOfBCnodes,
                                    real omega,
                                    unsigned int* neighborX,
                                    unsigned int* neighborY,
                                    unsigned int* neighborZ,
                                    real* turbViscosity,
                                    unsigned long long numberOfLBnodes,
                                    bool isEvenTimestep);

__global__ void QSlipGeomDeviceComp27(real* DD,
                                                 int* k_Q,
                                                 real* QQ,
                                                 unsigned int numberOfBCnodes,
                                                 real om1,
                                                 real* NormalX,
                                                 real* NormalY,
                                                 real* NormalZ,
                                                 unsigned int* neighborX,
                                                 unsigned int* neighborY,
                                                 unsigned int* neighborZ,
                                                 unsigned long long numberOfLBnodes,
                                                 bool isEvenTimestep);

__global__ void QSlipNormDeviceComp27(real* DD,
                                                 int* k_Q,
                                                 real* QQ,
                                                 unsigned int numberOfBCnodes,
                                                 real om1,
                                                 real* NormalX,
                                                 real* NormalY,
                                                 real* NormalZ,
                                                 unsigned int* neighborX,
                                                 unsigned int* neighborY,
                                                 unsigned int* neighborZ,
                                                 unsigned long long numberOfLBnodes,
                                                 bool isEvenTimestep);

__global__ void BBSlipDeviceComp27(
    real* distributions,
    int* subgridDistanceIndices,
    real* subgridDistances,
    unsigned int numberOfBCnodes,
    unsigned int* neighborX,
    unsigned int* neighborY,
    unsigned int* neighborZ,
    unsigned long long numberOfLBnodes,
    bool isEvenTimestep);

// Stress BCs (wall model)
__global__ void QStressDeviceComp27(real* DD,
                                               int* k_Q,
                                             int* k_N,
                                             real* QQ,
                                             unsigned int numberOfBCnodes,
                                             real om1,
                                             real* turbViscosity,
                                             real* vx,
                                             real* vy,
                                             real* vz,
                                             real* normalX,
                                             real* normalY,
                                             real* normalZ,
                                             real* vx_bc,
                                             real* vy_bc,
                                             real* vz_bc,
                                             real* vx1,
                                             real* vy1,
                                             real* vz1,
                                             int* samplingOffset,
                                             real* z0,
                                             bool  hasWallModelMonitor,
                                            real* u_star_monitor,
                                            real* Fx_monitor,
                                            real* Fy_monitor,
                                            real* Fz_monitor,
                                             unsigned int* neighborX,
                                             unsigned int* neighborY,
                                             unsigned int* neighborZ,
                                             unsigned long long numberOfLBnodes,
                                             bool isEvenTimestep);

__global__ void BBStressDevice27( real* DD,
                                                int* k_Q,
                                                int* k_N,
                                                real* QQ,
                                                unsigned int numberOfBCnodes,
                                                real* vx,
                                                real* vy,
                                                real* vz,
                                                real* normalX,
                                                real* normalY,
                                                real* normalZ,
                                                real* vx_bc,
                                                real* vy_bc,
                                                real* vz_bc,
                                                real* vx1,
                                                real* vy1,
                                                real* vz1,
                                                int* samplingOffset,
                                                real* z0,
                                                bool  hasWallModelMonitor,
                                                real* u_star_monitor,
                                                real* Fx_monitor,
                                                real* Fy_monitor,
                                                real* Fz_monitor,
                                                unsigned int* neighborX,
                                                unsigned int* neighborY,
                                                unsigned int* neighborZ,
                                                unsigned long long numberOfLBnodes,
                                                bool isEvenTimestep);

__global__ void BBStressPressureDevice27( real* DD,
                                                        int* k_Q,
                                             int* k_N,
                                             real* QQ,
                                             unsigned int  numberOfBCnodes,
                                             real* vx,
                                             real* vy,
                                             real* vz,
                                             real* normalX,
                                             real* normalY,
                                             real* normalZ,
                                             real* vx_el,
                                             real* vy_el,
                                             real* vz_el,
                                             real* vx_w_mean,
                                             real* vy_w_mean,
                                             real* vz_w_mean,
                                             int* samplingOffset,
                                             real* z0,
                                             bool  hasWallModelMonitor,
                                             real* u_star_monitor,
                                             real* Fx_monitor,
                                             real* Fy_monitor,
                                             real* Fz_monitor,
                                             unsigned int* neighborX,
                                             unsigned int* neighborY,
                                             unsigned int* neighborZ,
                                             unsigned long long numberOfLBnodes,
                                             bool isEvenTimestep);

//Pressure BCs
__global__ void QPressDevice27( real* rhoBC,
                                           real* DD,
                                           int* k_Q,
                                           real* QQ,
                                           unsigned int numberOfBCnodes,
                                           real om1,
                                           unsigned int* neighborX,
                                           unsigned int* neighborY,
                                           unsigned int* neighborZ,
                                           unsigned long long numberOfLBnodes,
                                           bool isEvenTimestep);

__global__ void QPressDeviceAntiBB27(   real* rhoBC,
                                                   real* vx,
                                                   real* vy,
                                                   real* vz,
                                                   real* DD,
                                                   int* k_Q,
                                                   real* QQ,
                                                   int numberOfBCnodes,
                                                   real om1,
                                                   unsigned int* neighborX,
                                                   unsigned int* neighborY,
                                                   unsigned int* neighborZ,
                                                   unsigned long long numberOfLBnodes,
                                                   bool isEvenTimestep);

__global__ void QPressDeviceFixBackflow27( real* rhoBC,
                                                      real* DD,
                                                      int* k_Q,
                                                      int numberOfBCnodes,
                                                      real om1,
                                                      unsigned int* neighborX,
                                                      unsigned int* neighborY,
                                                      unsigned int* neighborZ,
                                                      unsigned long long numberOfLBnodes,
                                                      bool isEvenTimestep);

__global__ void QPressDeviceDirDepBot27(  real* rhoBC,
                                                     real* DD,
                                                     int* k_Q,
                                                     int numberOfBCnodes,
                                                     real om1,
                                                     unsigned int* neighborX,
                                                     unsigned int* neighborY,
                                                     unsigned int* neighborZ,
                                                     unsigned long long numberOfLBnodes,
                                                     bool isEvenTimestep);

__global__ void QPressNoRhoDevice27(  real* rhoBC,
                                                 real* distributions,
                                                 int* k_Q,
                                                 int* k_N,
                                                 int numberOfBCnodes,
                                                 real om1,
                                                 unsigned int* neighborX,
                                                 unsigned int* neighborY,
                                                 unsigned int* neighborZ,
                                                 unsigned long long numberOfLBnodes,
                                                 bool isEvenTimestep,
                                                 int direction);

__global__ void QPressZeroRhoOutflowDevice27(  real* rhoBC,
                                            real* distributions,
                                            int* k_Q,
                                            int* k_N,
                                            int numberOfBCnodes,
                                            real om1,
                                            unsigned int* neighborX,
                                            unsigned int* neighborY,
                                            unsigned int* neighborZ,
                                            unsigned long long numberOfLBnodes,
                                            bool isEvenTimestep,
                                            int direction,
                                            real densityCorrectionFactor);

__global__ void QInflowScaleByPressDevice27(  real* rhoBC,
                                                         real* DD,
                                                         int* k_Q,
                                                         int* k_N,
                                                         int numberOfBCnodes,
                                                         real om1,
                                                         unsigned int* neighborX,
                                                         unsigned int* neighborY,
                                                         unsigned int* neighborZ,
                                                         unsigned long long numberOfLBnodes,
                                                         bool isEvenTimestep);

__global__ void QPressDeviceOld27(real* rhoBC,
                                             real* DD,
                                             int* k_Q,
                                             int* k_N,
                                             int numberOfBCnodes,
                                             real om1,
                                             unsigned int* neighborX,
                                             unsigned int* neighborY,
                                             unsigned int* neighborZ,
                                             unsigned long long numberOfLBnodes,
                                             bool isEvenTimestep);

__global__ void QPressDeviceIncompNEQ27( real* rhoBC,
                                                    real* DD,
                                                    int* k_Q,
                                                    int* k_N,
                                                    int numberOfBCnodes,
                                                    real om1,
                                                    unsigned int* neighborX,
                                                    unsigned int* neighborY,
                                                    unsigned int* neighborZ,
                                                    unsigned long long numberOfLBnodes,
                                                    bool isEvenTimestep);

__global__ void QPressDeviceNEQ27(real* rhoBC,
                                             real* distribution,
                                             int* bcNodeIndices,
                                             int* bcNeighborIndices,
                                             int numberOfBCnodes,
                                             real omega1,
                                             unsigned int* neighborX,
                                             unsigned int* neighborY,
                                             unsigned int* neighborZ,
                                             unsigned long long numberOfLBnodes,
                                             bool isEvenTimestep);

__global__ void QPressDeviceEQZ27(real* rhoBC,
                                             real* DD,
                                             int* k_Q,
                                             int* k_N,
                                             real* kTestRE,
                                             int numberOfBCnodes,
                                             real om1,
                                             unsigned int* neighborX,
                                             unsigned int* neighborY,
                                             unsigned int* neighborZ,
                                             unsigned long long numberOfLBnodes,
                                             bool isEvenTimestep);

__global__ void QPressDeviceZero27(  real* DD,
                                                int* k_Q,
                                                unsigned int numberOfBCnodes,
                                                unsigned int* neighborX,
                                                unsigned int* neighborY,
                                                unsigned int* neighborZ,
                                                unsigned long long numberOfLBnodes,
                                                bool isEvenTimestep);

__global__ void QPressDeviceFake27(real* rhoBC,
                                             real* DD,
                                             int* k_Q,
                                             int* k_N,
                                             int numberOfBCnodes,
                                             real om1,
                                             unsigned int* neighborX,
                                             unsigned int* neighborY,
                                             unsigned int* neighborZ,
                                             unsigned long long numberOfLBnodes,
                                             bool isEvenTimestep);

__global__ void BBDevice27(real* distributions,
                                     int* subgridDistanceIndices,
                                     real* subgridDistances,
                                     unsigned int numberOfBCnodes,
                                     unsigned int* neighborX,
                                     unsigned int* neighborY,
                                     unsigned int* neighborZ,
                                     unsigned long long numberOfLBnodes,
                                     bool isEvenTimestep);

__global__ void QPressDevice27_IntBB(real* rho,
                                                real* DD,
                                                int* k_Q,
                                                real* QQ,
                                                unsigned int numberOfBCnodes,
                                                real om1,
                                                unsigned int* neighborX,
                                                unsigned int* neighborY,
                                                unsigned int* neighborZ,
                                                unsigned long long numberOfLBnodes,
                                                bool isEvenTimestep);

// TODO: https://git.rz.tu-bs.de/irmb/VirtualFluids_dev/-/issues/29
//Schlaffer BCs
__global__ void PressSchlaff27(real* rhoBC,
                                          real* DD,
                                          real* vx0,
                                          real* vy0,
                                          real* vz0,
                                          real* deltaVz0,
                                          int* k_Q,
                                          int* k_N,
                                          int numberOfBCnodes,
                                          real om1,
                                          unsigned int* neighborX,
                                          unsigned int* neighborY,
                                          unsigned int* neighborZ,
                                          unsigned long long numberOfLBnodes,
                                          bool isEvenTimestep);

// TODO: https://git.rz.tu-bs.de/irmb/VirtualFluids_dev/-/issues/29
__global__ void VelSchlaff27(  int t,
                                          real* DD,
                                          real* vz0,
                                          real* deltaVz0,
                                          int* k_Q,
                                          int* k_N,
                                          int numberOfBCnodes,
                                          real om1,
                                          unsigned int* neighborX,
                                          unsigned int* neighborY,
                                          unsigned int* neighborZ,
                                          unsigned long long numberOfLBnodes,
                                          bool isEvenTimestep);

__global__ void QPrecursorDeviceCompZeroPress(     int* subgridDistanceIndices,
                                                int numberOfBCnodes,
                                                int numberOfPrecursorNodes,
                                                int sizeQ,
                                                real omega,
                                                real* distributions,
                                                real* subgridDistances,
                                                uint* neighborX,
                                                uint* neighborY,
                                                uint* neighborZ,
                                                uint* neighborsNT,
                                                uint* neighborsNB,
                                                uint* neighborsST,
                                                uint* neighborsSB,
                                                real* weights0PP,
                                                real* weights0PM,
                                                real* weights0MP,
                                                real* weights0MM,
                                                real* vLast,
                                                real* vCurrent,
                                                real velocityX,
                                                real velocityY,
                                                real velocityZ,
                                                real timeRatio,
                                                real velocityRatio,
                                                unsigned long long numberOfLBnodes,
                                                bool isEvenTimestep);

__global__ void PrecursorDeviceEQ27(     int* subgridDistanceIndices,
                                        int numberOfBCnodes,
                                        int numberOfPrecursorNodes,
                                        real omega,
                                        real* distributions,
                                        uint* neighborX,
                                        uint* neighborY,
                                        uint* neighborZ,
                                        uint* neighborsNT,
                                        uint* neighborsNB,
                                        uint* neighborsST,
                                        uint* neighborsSB,
                                        real* weights0PP,
                                        real* weights0PM,
                                        real* weights0MP,
                                        real* weights0MM,
                                        real* vLast,
                                        real* vCurrent,
                                        real velocityX,
                                        real velocityY,
                                        real velocityZ,
                                        real timeRatio,
                                        real velocityRatio,
                                        unsigned long long numberOfLBnodes,
                                        bool isEvenTimestep);

__global__ void PrecursorDeviceDistributions(     int* subgridDistanceIndices,
                                                int numberOfBCNodes,
                                                int numberOfPrecursorNodes,
                                                real* distributions,
                                                uint* neighborX,
                                                uint* neighborY,
                                                uint* neighborZ,
                                                uint* neighborsNT,
                                                uint* neighborsNB,
                                                uint* neighborsST,
                                                uint* neighborsSB,
                                                real* weights0PP,
                                                real* weights0PM,
                                                real* weights0MP,
                                                real* weights0MM,
                                                real* fsLast,
                                                real* fsNext,
                                                real timeRatio,
                                                unsigned long long numberOfLBnodes,
                                                bool isEvenTimestep);
__global__ void QPrecursorDeviceDistributions(     int* subgridDistanceIndices,
                                                real* subgridDistances,
                                                int sizeQ,
                                                int numberOfBCNodes,
                                                int numberOfPrecursorNodes,
                                                real* distributions,
                                                uint* neighborX,
                                                uint* neighborY,
                                                uint* neighborZ,
                                                uint* neighborsNT,
                                                uint* neighborsNB,
                                                uint* neighborsST,
                                                uint* neighborsSB,
                                                real* weights0PP,
                                                real* weights0PM,
                                                real* weights0MP,
                                                real* weights0MM,
                                                real* fsLast,
                                                real* fsNext,
                                                real timeRatio,
                                                unsigned long long numberOfLBnodes,
                                                bool isEvenTimestep);

//Advection / Diffusion BCs
__global__ void QAD7( real* DD,
                                 real* DD7,
                                 real* temp,
                                 real diffusivity,
                                 int* k_Q,
                                 real* QQ,
                                 unsigned int numberOfBCnodes,
                                 real om1,
                                 unsigned int* neighborX,
                                 unsigned int* neighborY,
                                 unsigned int* neighborZ,
                                 unsigned long long numberOfLBnodes,
                                 bool isEvenTimestep);

//////////////////////////////////////////////////////////////////////////
//! \brief \ref AD_SlipVelDeviceComp : device function for the slip-AD boundary condition
__global__ void AD_SlipVelDeviceComp(
    real * normalX,
    real * normalY,
    real * normalZ,
    real * distributions,
    real * distributionsAD,
    int* QindexArray,
    real * Qarrays,
    uint numberOfBCnodes,
    real omegaDiffusivity,
    uint * neighborX,
    uint * neighborY,
    uint * neighborZ,
    unsigned long long numberOfLBnodes,
    bool isEvenTimestep);

__global__ void QADDirichlet27(   real* DD,
                                             real* DD27,
                                             real* temp,
                                             real diffusivity,
                                             int* k_Q,
                                             real* QQ,
                                             unsigned int numberOfBCnodes,
                                             real om1,
                                             unsigned int* neighborX,
                                             unsigned int* neighborY,
                                             unsigned int* neighborZ,
                                             unsigned long long numberOfLBnodes,
                                             bool isEvenTimestep);

__global__ void QADBB27(  real* DD,
                                     real* DD27,
                                     real* temp,
                                     real diffusivity,
                                     int* k_Q,
                                     real* QQ,
                                     unsigned int numberOfBCnodes,
                                     real om1,
                                     unsigned int* neighborX,
                                     unsigned int* neighborY,
                                     unsigned int* neighborZ,
                                     unsigned long long numberOfLBnodes,
                                     bool isEvenTimestep);

__global__ void QADVel7( real* DD,
                                    real* DD7,
                                    real* temp,
                                    real* velo,
                                    real diffusivity,
                                    int* k_Q,
                                    real* QQ,
                                    unsigned int numberOfBCnodes,
                                    real om1,
                                    unsigned int* neighborX,
                                    unsigned int* neighborY,
                                    unsigned int* neighborZ,
                                    unsigned long long numberOfLBnodes,
                                    bool isEvenTimestep);

__global__ void QADVel27(real* DD,
                                    real* DD27,
                                    real* temp,
                                    real* velo,
                                    real diffusivity,
                                    int* k_Q,
                                    real* QQ,
                                    unsigned int numberOfBCnodes,
                                    real om1,
                                    unsigned int* neighborX,
                                    unsigned int* neighborY,
                                    unsigned int* neighborZ,
                                    unsigned long long numberOfLBnodes,
                                    bool isEvenTimestep);

__global__ void QADPress7(  real* DD,
                                       real* DD7,
                                       real* temp,
                                       real* velo,
                                       real diffusivity,
                                       int* k_Q,
                                       real* QQ,
                                       unsigned int numberOfBCnodes,
                                       real om1,
                                       unsigned int* neighborX,
                                       unsigned int* neighborY,
                                       unsigned int* neighborZ,
                                       unsigned long long numberOfLBnodes,
                                       bool isEvenTimestep);

__global__ void QADPress27( real* DD,
                                       real* DD27,
                                       real* temp,
                                       real* velo,
                                       real diffusivity,
                                       int* k_Q,
                                       real* QQ,
                                       unsigned int numberOfBCnodes,
                                       real om1,
                                       unsigned int* neighborX,
                                       unsigned int* neighborY,
                                       unsigned int* neighborZ,
                                       unsigned long long numberOfLBnodes,
                                       bool isEvenTimestep);

__global__ void QADPressNEQNeighbor27(
                                                 real* DD,
                                                 real* DD27,
                                                 int* k_Q,
                                                 int* k_N,
                                                 int numberOfBCnodes,
                                                 unsigned int* neighborX,
                                                 unsigned int* neighborY,
                                                 unsigned int* neighborZ,
                                                 unsigned long long numberOfLBnodes,
                                                 bool isEvenTimestep
                                                );

__global__ void QNoSlipADincomp7( real* DD,
                                             real* DD7,
                                             real* temp,
                                             real diffusivity,
                                             int* k_Q,
                                             real* QQ,
                                             unsigned int numberOfBCnodes,
                                             real om1,
                                             unsigned int* neighborX,
                                             unsigned int* neighborY,
                                             unsigned int* neighborZ,
                                             unsigned long long numberOfLBnodes,
                                             bool isEvenTimestep);

__global__ void QNoSlipADincomp27( real* DD,
                                             real* DD27,
                                             real* temp,
                                             real diffusivity,
                                             int* k_Q,
                                             real* QQ,
                                             unsigned int numberOfBCnodes,
                                             real om1,
                                             unsigned int* neighborX,
                                             unsigned int* neighborY,
                                             unsigned int* neighborZ,
                                             unsigned long long numberOfLBnodes,
                                             bool isEvenTimestep);

__global__ void QADVeloIncomp7(  real* DD,
                                            real* DD7,
                                            real* temp,
                                            real* velo,
                                            real diffusivity,
                                            int* k_Q,
                                            real* QQ,
                                            unsigned int numberOfBCnodes,
                                            real om1,
                                            unsigned int* neighborX,
                                            unsigned int* neighborY,
                                            unsigned int* neighborZ,
                                            unsigned long long numberOfLBnodes,
                                            bool isEvenTimestep);

__global__ void QADVeloIncomp27( real* DD,
                                            real* DD27,
                                            real* temp,
                                            real* velo,
                                            real diffusivity,
                                            int* k_Q,
                                            real* QQ,
                                            unsigned int numberOfBCnodes,
                                            real om1,
                                            unsigned int* neighborX,
                                            unsigned int* neighborY,
                                            unsigned int* neighborZ,
                                            unsigned long long numberOfLBnodes,
                                            bool isEvenTimestep);

__global__ void QADPressIncomp7(real* DD,
                                           real* DD7,
                                           real* temp,
                                           real* velo,
                                           real diffusivity,
                                           int* k_Q,
                                           real* QQ,
                                           unsigned int numberOfBCnodes,
                                           real om1,
                                           unsigned int* neighborX,
                                           unsigned int* neighborY,
                                           unsigned int* neighborZ,
                                           unsigned long long numberOfLBnodes,
                                           bool isEvenTimestep);

__global__ void QADPressIncomp27(   real* DD,
                                               real* DD27,
                                               real* temp,
                                               real* velo,
                                               real diffusivity,
                                               int* k_Q,
                                               real* QQ,
                                               unsigned int numberOfBCnodes,
                                               real om1,
                                               unsigned int* neighborX,
                                               unsigned int* neighborY,
                                               unsigned int* neighborZ,
                                               unsigned long long numberOfLBnodes,
                                               bool isEvenTimestep);



//coarse to fine
__global__ void scaleCF27(real* DC,
                                     real* DF,
                                    unsigned int* neighborCX,
                                    unsigned int* neighborCY,
                                    unsigned int* neighborCZ,
                                    unsigned int* neighborFX,
                                    unsigned int* neighborFY,
                                    unsigned int* neighborFZ,
                                               unsigned long long numberOfLBnodesC,
                                               unsigned long long numberOfLBnodesF,
                                               bool isEvenTimestep,
                                     unsigned int* posCSWB,
                                     unsigned int* posFSWB,
                                     unsigned int kCF,
                                               real omCoarse,
                                               real omFine,
                                               real nu,
                                               unsigned int nxC,
                                               unsigned int nyC,
                                               unsigned int nxF,
                                               unsigned int nyF);

__global__ void scaleCFEff27(real* DC,
                                        real* DF,
                                        unsigned int* neighborCX,
                                        unsigned int* neighborCY,
                                        unsigned int* neighborCZ,
                                        unsigned int* neighborFX,
                                        unsigned int* neighborFY,
                                        unsigned int* neighborFZ,
                                                 unsigned long long numberOfLBnodesC,
                                                 unsigned long long numberOfLBnodesF,
                                                 bool isEvenTimestep,
                                        unsigned int* posCSWB,
                                        unsigned int* posFSWB,
                                        unsigned int kCF,
                                                 real omCoarse,
                                                 real omFine,
                                                 real nu,
                                                 unsigned int nxC,
                                                 unsigned int nyC,
                                                 unsigned int nxF,
                                        unsigned int nyF,
                                        ICellNeigh neighborCoarseToFine);

__global__ void scaleCFLast27( real* DC,
                                          real* DF,
                                          unsigned int* neighborCX,
                                          unsigned int* neighborCY,
                                          unsigned int* neighborCZ,
                                          unsigned int* neighborFX,
                                          unsigned int* neighborFY,
                                          unsigned int* neighborFZ,
                                          unsigned long long numberOfLBnodesC,
                                          unsigned long long numberOfLBnodesF,
                                          bool isEvenTimestep,
                                          unsigned int* posCSWB,
                                          unsigned int* posFSWB,
                                          unsigned int kCF,
                                          real omCoarse,
                                          real omFine,
                                          real nu,
                                          unsigned int nxC,
                                          unsigned int nyC,
                                          unsigned int nxF,
                                          unsigned int nyF,
                                          ICellNeigh neighborCoarseToFine);

__global__ void scaleCFpress27(real* DC,
                                          real* DF,
                                          unsigned int* neighborCX,
                                          unsigned int* neighborCY,
                                          unsigned int* neighborCZ,
                                          unsigned int* neighborFX,
                                          unsigned int* neighborFY,
                                          unsigned int* neighborFZ,
                                          unsigned long long numberOfLBnodesC,
                                          unsigned long long numberOfLBnodesF,
                                          bool isEvenTimestep,
                                          unsigned int* posCSWB,
                                          unsigned int* posFSWB,
                                          unsigned int kCF,
                                          real omCoarse,
                                          real omFine,
                                          real nu,
                                          unsigned int nxC,
                                          unsigned int nyC,
                                          unsigned int nxF,
                                          unsigned int nyF,
                                          ICellNeigh neighborCoarseToFine);

__global__ void scaleCF_Fix_27(real* DC,
                                          real* DF,
                                          unsigned int* neighborCX,
                                          unsigned int* neighborCY,
                                          unsigned int* neighborCZ,
                                          unsigned int* neighborFX,
                                          unsigned int* neighborFY,
                                          unsigned int* neighborFZ,
                                          unsigned long long numberOfLBnodesC,
                                          unsigned long long numberOfLBnodesF,
                                          bool isEvenTimestep,
                                          unsigned int* posCSWB,
                                          unsigned int* posFSWB,
                                          unsigned int kCF,
                                          real omCoarse,
                                          real omFine,
                                          real nu,
                                          unsigned int nxC,
                                          unsigned int nyC,
                                          unsigned int nxF,
                                          unsigned int nyF,
                                          ICellNeigh neighborCoarseToFine);

__global__ void scaleCF_Fix_comp_27(   real* DC,
                                                  real* DF,
                                                  unsigned int* neighborCX,
                                                  unsigned int* neighborCY,
                                                  unsigned int* neighborCZ,
                                                  unsigned int* neighborFX,
                                                  unsigned int* neighborFY,
                                                  unsigned int* neighborFZ,
                                                  unsigned long long numberOfLBnodesC,
                                                  unsigned long long numberOfLBnodesF,
                                                  bool isEvenTimestep,
                                                  unsigned int* posCSWB,
                                                  unsigned int* posFSWB,
                                                  unsigned int kCF,
                                                  real omCoarse,
                                                  real omFine,
                                                  real nu,
                                                  unsigned int nxC,
                                                  unsigned int nyC,
                                                  unsigned int nxF,
                                                  unsigned int nyF,
                                                  ICellNeigh neighborCoarseToFine);

__global__ void scaleCF_0817_comp_27(  real* DC,
                                                  real* DF,
                                                  unsigned int* neighborCX,
                                                  unsigned int* neighborCY,
                                                  unsigned int* neighborCZ,
                                                  unsigned int* neighborFX,
                                                  unsigned int* neighborFY,
                                                  unsigned int* neighborFZ,
                                                  unsigned long long numberOfLBnodesC,
                                                  unsigned long long numberOfLBnodesF,
                                                  bool isEvenTimestep,
                                                  unsigned int* posCSWB,
                                                  unsigned int* posFSWB,
                                                  unsigned int kCF,
                                                  real omCoarse,
                                                  real omFine,
                                                  real nu,
                                                  unsigned int nxC,
                                                  unsigned int nyC,
                                                  unsigned int nxF,
                                                  unsigned int nyF,
                                                  ICellNeigh neighborCoarseToFine);

__global__ void scaleCF_comp_D3Q27F3_2018( real* DC,
                                                      real* DF,
                                                      real* G6,
                                                      unsigned int* neighborCX,
                                                      unsigned int* neighborCY,
                                                      unsigned int* neighborCZ,
                                                      unsigned int* neighborFX,
                                                      unsigned int* neighborFY,
                                                      unsigned int* neighborFZ,
                                                      unsigned long long numberOfLBnodesC,
                                                      unsigned long long numberOfLBnodesF,
                                                      bool isEvenTimestep,
                                                      unsigned int* posCSWB,
                                                      unsigned int* posFSWB,
                                                      unsigned int kCF,
                                                      real omCoarse,
                                                      real omFine,
                                                      real nu,
                                                      unsigned int nxC,
                                                      unsigned int nyC,
                                                      unsigned int nxF,
                                                      unsigned int nyF,
                                                      ICellNeigh neighborCoarseToFine);

__global__ void scaleCF_comp_D3Q27F3( real* DC,
                                                 real* DF,
                                                 real* G6,
                                                 unsigned int* neighborCX,
                                                 unsigned int* neighborCY,
                                                 unsigned int* neighborCZ,
                                                 unsigned int* neighborFX,
                                                 unsigned int* neighborFY,
                                                 unsigned int* neighborFZ,
                                                 unsigned long long numberOfLBnodesC,
                                                 unsigned long long numberOfLBnodesF,
                                                 bool isEvenTimestep,
                                                 unsigned int* posCSWB,
                                                 unsigned int* posFSWB,
                                                 unsigned int kCF,
                                                 real omCoarse,
                                                 real omFine,
                                                 real nu,
                                                 unsigned int nxC,
                                                 unsigned int nyC,
                                                 unsigned int nxF,
                                                 unsigned int nyF,
                                                 ICellNeigh neighborCoarseToFine);


__global__ void scaleCF_staggered_time_comp_27(real* DC,
                                                          real* DF,
                                                          unsigned int* neighborCX,
                                                          unsigned int* neighborCY,
                                                          unsigned int* neighborCZ,
                                                          unsigned int* neighborFX,
                                                          unsigned int* neighborFY,
                                                          unsigned int* neighborFZ,
                                                          unsigned long long numberOfLBnodesC,
                                                          unsigned long long numberOfLBnodesF,
                                                          bool isEvenTimestep,
                                                          unsigned int* posCSWB,
                                                          unsigned int* posFSWB,
                                                          unsigned int kCF,
                                                          real omCoarse,
                                                          real omFine,
                                                          real nu,
                                                          unsigned int nxC,
                                                          unsigned int nyC,
                                                          unsigned int nxF,
                                                          unsigned int nyF,
                                                          ICellNeigh neighborCoarseToFine);

__global__ void scaleCF_RhoSq_comp_27( real* DC,
                                                  real* DF,
                                                  unsigned int* neighborCX,
                                                  unsigned int* neighborCY,
                                                  unsigned int* neighborCZ,
                                                  unsigned int* neighborFX,
                                                  unsigned int* neighborFY,
                                                  unsigned int* neighborFZ,
                                                  unsigned long long numberOfLBnodesC,
                                                  unsigned long long numberOfLBnodesF,
                                                  bool isEvenTimestep,
                                                  unsigned int* posCSWB,
                                                  unsigned int* posFSWB,
                                                  unsigned int kCF,
                                                  real omCoarse,
                                                  real omFine,
                                                  real nu,
                                                  unsigned int nxC,
                                                  unsigned int nyC,
                                                  unsigned int nxF,
                                                  unsigned int nyF,
                                                  ICellNeigh neighborCoarseToFine);

template<bool hasTurbulentViscosity> __global__ void scaleCF_compressible(
    real* distributionsCoarse,
    real* distributionsFine,
    unsigned int* neighborXcoarse,
    unsigned int* neighborYcoarse,
    unsigned int* neighborZcoarse,
    unsigned int* neighborXfine,
    unsigned int* neighborYfine,
    unsigned int* neighborZfine,
    unsigned long long numberOfLBnodesCoarse,
    unsigned long long numberOfLBnodesFine,
    bool isEvenTimestep,
    unsigned int* indicesCoarseMMM,
    unsigned int* indicesFineMMM,
    unsigned int numberOfInterfaceNodes,
    real omegaCoarse,
    real omegaFine,
    real* turbulentViscosityCoarse,
    real* turbulentViscosityFine,
    ICellNeigh offsetCF);

__global__ void scaleCF_RhoSq_3rdMom_comp_27(real* DC,
                                                        real* DF,
                                                        unsigned int* neighborCX,
                                                        unsigned int* neighborCY,
                                                        unsigned int* neighborCZ,
                                                        unsigned int* neighborFX,
                                                        unsigned int* neighborFY,
                                                        unsigned int* neighborFZ,
                                                        unsigned long long numberOfLBnodesC,
                                                        unsigned long long numberOfLBnodesF,
                                                        bool isEvenTimestep,
                                                        unsigned int* posCSWB,
                                                        unsigned int* posFSWB,
                                                        unsigned int kCF,
                                                        real omCoarse,
                                                        real omFine,
                                                        real nu,
                                                        unsigned int nxC,
                                                        unsigned int nyC,
                                                        unsigned int nxF,
                                                        unsigned int nyF,
                                                        ICellNeigh neighborCoarseToFine);

__global__ void scaleCF_AA2016_comp_27(real* DC,
                                                  real* DF,
                                                  unsigned int* neighborCX,
                                                  unsigned int* neighborCY,
                                                  unsigned int* neighborCZ,
                                                  unsigned int* neighborFX,
                                                  unsigned int* neighborFY,
                                                  unsigned int* neighborFZ,
                                                  unsigned long long numberOfLBnodesC,
                                                  unsigned long long numberOfLBnodesF,
                                                  bool isEvenTimestep,
                                                  unsigned int* posCSWB,
                                                  unsigned int* posFSWB,
                                                  unsigned int kCF,
                                                  real omCoarse,
                                                  real omFine,
                                                  real nu,
                                                  unsigned int nxC,
                                                  unsigned int nyC,
                                                  unsigned int nxF,
                                                  unsigned int nyF,
                                                  ICellNeigh neighborCoarseToFine);

__global__ void scaleCF_NSPress_27(real* DC,
                                              real* DF,
                                              unsigned int* neighborCX,
                                              unsigned int* neighborCY,
                                              unsigned int* neighborCZ,
                                              unsigned int* neighborFX,
                                              unsigned int* neighborFY,
                                              unsigned int* neighborFZ,
                                              unsigned long long numberOfLBnodesC,
                                              unsigned long long numberOfLBnodesF,
                                              bool isEvenTimestep,
                                              unsigned int* posCSWB,
                                              unsigned int* posFSWB,
                                              unsigned int kCF,
                                              real omCoarse,
                                              real omFine,
                                              real nu,
                                              unsigned int nxC,
                                              unsigned int nyC,
                                              unsigned int nxF,
                                              unsigned int nyF,
                                              ICellNeigh neighborCoarseToFine);

__global__ void scaleCFThSMG7( real* DC,
                                          real* DF,
                                          real* DD7C,
                                          real* DD7F,
                                          unsigned int* neighborCX,
                                          unsigned int* neighborCY,
                                          unsigned int* neighborCZ,
                                          unsigned int* neighborFX,
                                          unsigned int* neighborFY,
                                          unsigned int* neighborFZ,
                                          unsigned long long numberOfLBnodesC,
                                          unsigned long long numberOfLBnodesF,
                                          bool isEvenTimestep,
                                          unsigned int* posCSWB,
                                          unsigned int* posFSWB,
                                          unsigned int kCF,
                                          real nu,
                                          real diffusivity_fine,
                                          ICellNeigh neighborCoarseToFine);

__global__ void scaleCFThS7(real* DC,
                                       real* DF,
                                       real* DD7C,
                                       real* DD7F,
                                       unsigned int* neighborCX,
                                       unsigned int* neighborCY,
                                       unsigned int* neighborCZ,
                                       unsigned int* neighborFX,
                                       unsigned int* neighborFY,
                                       unsigned int* neighborFZ,
                                       unsigned long long numberOfLBnodesC,
                                       unsigned long long numberOfLBnodesF,
                                       bool isEvenTimestep,
                                       unsigned int* posCSWB,
                                       unsigned int* posFSWB,
                                       unsigned int kCF,
                                       real nu,
                                       real diffusivity_fine);

__global__ void scaleCFThS27(real* DC,
                                        real* DF,
                                        real* DD27C,
                                        real* DD27F,
                                        unsigned int* neighborCX,
                                        unsigned int* neighborCY,
                                        unsigned int* neighborCZ,
                                        unsigned int* neighborFX,
                                        unsigned int* neighborFY,
                                        unsigned int* neighborFZ,
                                        unsigned long long numberOfLBnodesC,
                                        unsigned long long numberOfLBnodesF,
                                        bool isEvenTimestep,
                                        unsigned int* posCSWB,
                                        unsigned int* posFSWB,
                                        unsigned int kCF,
                                        real nu,
                                        real diffusivity_fine,
                                        ICellNeigh neighborCoarseToFine);

//fine to coarse
__global__ void scaleFC27(real* DC,
                                     real* DF,
                                    unsigned int* neighborCX,
                                    unsigned int* neighborCY,
                                    unsigned int* neighborCZ,
                                    unsigned int* neighborFX,
                                    unsigned int* neighborFY,
                                    unsigned int* neighborFZ,
                                               unsigned long long numberOfLBnodesC,
                                               unsigned long long numberOfLBnodesF,
                                               bool isEvenTimestep,
                                     unsigned int* posC,
                                     unsigned int* posFSWB,
                                     unsigned int kFC,
                                               real omCoarse,
                                               real omFine,
                                               real nu,
                                               unsigned int nxC,
                                               unsigned int nyC,
                                               unsigned int nxF,
                                     unsigned int nyF);

__global__ void scaleFCEff27(real* DC,
                                        real* DF,
                                        unsigned int* neighborCX,
                                        unsigned int* neighborCY,
                                        unsigned int* neighborCZ,
                                        unsigned int* neighborFX,
                                        unsigned int* neighborFY,
                                        unsigned int* neighborFZ,
                                        unsigned long long numberOfLBnodesC,
                                        unsigned long long numberOfLBnodesF,
                                        bool isEvenTimestep,
                                        unsigned int* posC,
                                        unsigned int* posFSWB,
                                        unsigned int kFC,
                                        real omCoarse,
                                        real omFine,
                                        real nu,
                                        unsigned int nxC,
                                        unsigned int nyC,
                                        unsigned int nxF,
                                        unsigned int nyF,
                                        ICellNeigh neighborFineToCoarse);

__global__ void scaleFCLast27( real* DC,
                                          real* DF,
                                          unsigned int* neighborCX,
                                          unsigned int* neighborCY,
                                          unsigned int* neighborCZ,
                                          unsigned int* neighborFX,
                                          unsigned int* neighborFY,
                                          unsigned int* neighborFZ,
                                          unsigned long long numberOfLBnodesC,
                                          unsigned long long numberOfLBnodesF,
                                          bool isEvenTimestep,
                                          unsigned int* posC,
                                          unsigned int* posFSWB,
                                          unsigned int kFC,
                                          real omCoarse,
                                          real omFine,
                                          real nu,
                                          unsigned int nxC,
                                          unsigned int nyC,
                                          unsigned int nxF,
                                          unsigned int nyF,
                                          ICellNeigh neighborFineToCoarse);

__global__ void scaleFCpress27( real* DC,
                                          real* DF,
                                          unsigned int* neighborCX,
                                          unsigned int* neighborCY,
                                          unsigned int* neighborCZ,
                                          unsigned int* neighborFX,
                                          unsigned int* neighborFY,
                                          unsigned int* neighborFZ,
                                          unsigned long long numberOfLBnodesC,
                                          unsigned long long numberOfLBnodesF,
                                          bool isEvenTimestep,
                                          unsigned int* posC,
                                          unsigned int* posFSWB,
                                          unsigned int kFC,
                                          real omCoarse,
                                          real omFine,
                                          real nu,
                                          unsigned int nxC,
                                          unsigned int nyC,
                                          unsigned int nxF,
                                          unsigned int nyF,
                                          ICellNeigh neighborFineToCoarse);

__global__ void scaleFC_Fix_27( real* DC,
                                          real* DF,
                                          unsigned int* neighborCX,
                                          unsigned int* neighborCY,
                                          unsigned int* neighborCZ,
                                          unsigned int* neighborFX,
                                          unsigned int* neighborFY,
                                          unsigned int* neighborFZ,
                                          unsigned long long numberOfLBnodesC,
                                          unsigned long long numberOfLBnodesF,
                                          bool isEvenTimestep,
                                          unsigned int* posC,
                                          unsigned int* posFSWB,
                                          unsigned int kFC,
                                          real omCoarse,
                                          real omFine,
                                          real nu,
                                          unsigned int nxC,
                                          unsigned int nyC,
                                          unsigned int nxF,
                                          unsigned int nyF,
                                          ICellNeigh neighborFineToCoarse);

__global__ void scaleFC_Fix_comp_27(   real* DC,
                                                  real* DF,
                                                  unsigned int* neighborCX,
                                                  unsigned int* neighborCY,
                                                  unsigned int* neighborCZ,
                                                  unsigned int* neighborFX,
                                                  unsigned int* neighborFY,
                                                  unsigned int* neighborFZ,
                                                  unsigned long long numberOfLBnodesC,
                                                  unsigned long long numberOfLBnodesF,
                                                  bool isEvenTimestep,
                                                  unsigned int* posC,
                                                  unsigned int* posFSWB,
                                                  unsigned int kFC,
                                                  real omCoarse,
                                                  real omFine,
                                                  real nu,
                                                  unsigned int nxC,
                                                  unsigned int nyC,
                                                  unsigned int nxF,
                                                  unsigned int nyF,
                                                  ICellNeigh neighborFineToCoarse);

__global__ void scaleFC_0817_comp_27(  real* DC,
                                                  real* DF,
                                                  unsigned int* neighborCX,
                                                  unsigned int* neighborCY,
                                                  unsigned int* neighborCZ,
                                                  unsigned int* neighborFX,
                                                  unsigned int* neighborFY,
                                                  unsigned int* neighborFZ,
                                                  unsigned long long numberOfLBnodesC,
                                                  unsigned long long numberOfLBnodesF,
                                                  bool isEvenTimestep,
                                                  unsigned int* posC,
                                                  unsigned int* posFSWB,
                                                  unsigned int kFC,
                                                  real omCoarse,
                                                  real omFine,
                                                  real nu,
                                                  unsigned int nxC,
                                                  unsigned int nyC,
                                                  unsigned int nxF,
                                                  unsigned int nyF,
                                                  ICellNeigh neighborFineToCoarse);

__global__ void scaleFC_comp_D3Q27F3_2018( real* DC,
                                                      real* DF,
                                                      real* G6,
                                                      unsigned int* neighborCX,
                                                      unsigned int* neighborCY,
                                                      unsigned int* neighborCZ,
                                                      unsigned int* neighborFX,
                                                      unsigned int* neighborFY,
                                                      unsigned int* neighborFZ,
                                                      unsigned long long numberOfLBnodesC,
                                                      unsigned long long numberOfLBnodesF,
                                                      bool isEvenTimestep,
                                                      unsigned int* posC,
                                                      unsigned int* posFSWB,
                                                      unsigned int kFC,
                                                      real omCoarse,
                                                      real omFine,
                                                      real nu,
                                                      unsigned int nxC,
                                                      unsigned int nyC,
                                                      unsigned int nxF,
                                                      unsigned int nyF,
                                                      ICellNeigh neighborFineToCoarse);

__global__ void scaleFC_comp_D3Q27F3( real* DC,
                                                 real* DF,
                                                 real* G6,
                                                 unsigned int* neighborCX,
                                                 unsigned int* neighborCY,
                                                 unsigned int* neighborCZ,
                                                 unsigned int* neighborFX,
                                                 unsigned int* neighborFY,
                                                 unsigned int* neighborFZ,
                                                 unsigned long long numberOfLBnodesC,
                                                 unsigned long long numberOfLBnodesF,
                                                 bool isEvenTimestep,
                                                 unsigned int* posC,
                                                 unsigned int* posFSWB,
                                                 unsigned int kFC,
                                                 real omCoarse,
                                                 real omFine,
                                                 real nu,
                                                 unsigned int nxC,
                                                 unsigned int nyC,
                                                 unsigned int nxF,
                                                 unsigned int nyF,
                                                 ICellNeigh neighborFineToCoarse);


__global__ void scaleFC_staggered_time_comp_27(real* DC,
                                                          real* DF,
                                                          unsigned int* neighborCX,
                                                          unsigned int* neighborCY,
                                                          unsigned int* neighborCZ,
                                                          unsigned int* neighborFX,
                                                          unsigned int* neighborFY,
                                                          unsigned int* neighborFZ,
                                                          unsigned long long numberOfLBnodesC,
                                                          unsigned long long numberOfLBnodesF,
                                                          bool isEvenTimestep,
                                                          unsigned int* posC,
                                                          unsigned int* posFSWB,
                                                          unsigned int kFC,
                                                          real omCoarse,
                                                          real omFine,
                                                          real nu,
                                                          unsigned int nxC,
                                                          unsigned int nyC,
                                                          unsigned int nxF,
                                                          unsigned int nyF,
                                                          ICellNeigh neighborFineToCoarse);

__global__ void scaleFC_RhoSq_comp_27( real* DC,
                                                  real* DF,
                                                  unsigned int* neighborCX,
                                                  unsigned int* neighborCY,
                                                  unsigned int* neighborCZ,
                                                  unsigned int* neighborFX,
                                                  unsigned int* neighborFY,
                                                  unsigned int* neighborFZ,
                                                  unsigned long long numberOfLBnodesC,
                                                  unsigned long long numberOfLBnodesF,
                                                  bool isEvenTimestep,
                                                  unsigned int* posC,
                                                  unsigned int* posFSWB,
                                                  unsigned int kFC,
                                                  real omCoarse,
                                                  real omFine,
                                                  real nu,
                                                  unsigned int nxC,
                                                  unsigned int nyC,
                                                  unsigned int nxF,
                                                  unsigned int nyF,
                                                  ICellNeigh neighborFineToCoarse);

template<bool hasTurbulentViscosity> __global__ void scaleFC_compressible(
    real *distributionsCoarse,
    real *distributionsFine,
    unsigned int *neighborXcoarse,
    unsigned int *neighborYcoarse,
    unsigned int *neighborZcoarse,
    unsigned int *neighborXfine,
    unsigned int *neighborYfine,
    unsigned int *neighborZfine,
    unsigned long long numberOfLBnodesCoarse,
    unsigned long long numberOfLBnodesFine,
    bool isEvenTimestep,
    unsigned int *indicesCoarse000,
    unsigned int *indicesFineMMM,
    unsigned int numberOfInterfaceNodes,
    real omegaCoarse,
    real omegaFine,
    real* turbulentViscosityCoarse,
    real* turbulentViscosityFine,
    ICellNeigh offsetFC);

__global__ void scaleFC_RhoSq_3rdMom_comp_27(real* DC,
                                                        real* DF,
                                                        unsigned int* neighborCX,
                                                        unsigned int* neighborCY,
                                                        unsigned int* neighborCZ,
                                                        unsigned int* neighborFX,
                                                        unsigned int* neighborFY,
                                                        unsigned int* neighborFZ,
                                                        unsigned long long numberOfLBnodesC,
                                                        unsigned long long numberOfLBnodesF,
                                                        bool isEvenTimestep,
                                                        unsigned int* posC,
                                                        unsigned int* posFSWB,
                                                        unsigned int kFC,
                                                        real omCoarse,
                                                        real omFine,
                                                        real nu,
                                                        unsigned int nxC,
                                                        unsigned int nyC,
                                                        unsigned int nxF,
                                                        unsigned int nyF,
                                                        ICellNeigh neighborFineToCoarse);

__global__ void scaleFC_AA2016_comp_27(real* DC,
                                                  real* DF,
                                                  unsigned int* neighborCX,
                                                  unsigned int* neighborCY,
                                                  unsigned int* neighborCZ,
                                                  unsigned int* neighborFX,
                                                  unsigned int* neighborFY,
                                                  unsigned int* neighborFZ,
                                                  unsigned long long numberOfLBnodesC,
                                                  unsigned long long numberOfLBnodesF,
                                                  bool isEvenTimestep,
                                                  unsigned int* posC,
                                                  unsigned int* posFSWB,
                                                  unsigned int kFC,
                                                  real omCoarse,
                                                  real omFine,
                                                  real nu,
                                                  unsigned int nxC,
                                                  unsigned int nyC,
                                                  unsigned int nxF,
                                                  unsigned int nyF,
                                                  ICellNeigh neighborFineToCoarse);

__global__ void scaleFC_NSPress_27(real* DC,
                                              real* DF,
                                              unsigned int* neighborCX,
                                              unsigned int* neighborCY,
                                              unsigned int* neighborCZ,
                                              unsigned int* neighborFX,
                                              unsigned int* neighborFY,
                                              unsigned int* neighborFZ,
                                              unsigned long long numberOfLBnodesC,
                                              unsigned long long numberOfLBnodesF,
                                              bool isEvenTimestep,
                                              unsigned int* posC,
                                              unsigned int* posFSWB,
                                              unsigned int kFC,
                                              real omCoarse,
                                              real omFine,
                                              real nu,
                                              unsigned int nxC,
                                              unsigned int nyC,
                                              unsigned int nxF,
                                              unsigned int nyF,
                                              ICellNeigh neighborFineToCoarse);

__global__ void scaleFCThSMG7( real* DC,
                                          real* DF,
                                          real* DD7C,
                                          real* DD7F,
                                          unsigned int* neighborCX,
                                          unsigned int* neighborCY,
                                          unsigned int* neighborCZ,
                                          unsigned int* neighborFX,
                                          unsigned int* neighborFY,
                                          unsigned int* neighborFZ,
                                          unsigned long long numberOfLBnodesC,
                                          unsigned long long numberOfLBnodesF,
                                          bool isEvenTimestep,
                                          unsigned int* posC,
                                          unsigned int* posFSWB,
                                          unsigned int kFC,
                                          real nu,
                                          real diffusivity_coarse,
                                          ICellNeigh neighborFineToCoarse);

__global__ void scaleFCThS7(real* DC,
                                       real* DF,
                                       real* DD7C,
                                       real* DD7F,
                                       unsigned int* neighborCX,
                                       unsigned int* neighborCY,
                                       unsigned int* neighborCZ,
                                       unsigned int* neighborFX,
                                       unsigned int* neighborFY,
                                       unsigned int* neighborFZ,
                                       unsigned long long numberOfLBnodesC,
                                       unsigned long long numberOfLBnodesF,
                                       bool isEvenTimestep,
                                       unsigned int* posC,
                                       unsigned int* posFSWB,
                                       unsigned int kFC,
                                       real nu,
                                       real diffusivity_coarse);

__global__ void scaleFCThS27(  real* DC,
                                          real* DF,
                                          real* DD27C,
                                          real* DD27F,
                                          unsigned int* neighborCX,
                                          unsigned int* neighborCY,
                                          unsigned int* neighborCZ,
                                          unsigned int* neighborFX,
                                          unsigned int* neighborFY,
                                          unsigned int* neighborFZ,
                                          unsigned long long numberOfLBnodesC,
                                          unsigned long long numberOfLBnodesF,
                                          bool isEvenTimestep,
                                          unsigned int* posC,
                                          unsigned int* posFSWB,
                                          unsigned int kFC,
                                          real nu,
                                          real diffusivity_coarse,
                                          ICellNeigh neighborFineToCoarse);

__global__ void DragLiftPost27(  real* DD,
                                            int* k_Q,
                                            real* QQ,
                                            int numberOfBCnodes,
                                            double *DragX,
                                            double *DragY,
                                            double *DragZ,
                                            unsigned int* neighborX,
                                            unsigned int* neighborY,
                                            unsigned int* neighborZ,
                                            unsigned long long numberOfLBnodes,
                                            bool isEvenTimestep);

__global__ void DragLiftPre27(   real* DD,
                                            int* k_Q,
                                            real* QQ,
                                            int numberOfBCnodes,
                                            double *DragX,
                                            double *DragY,
                                            double *DragZ,
                                            unsigned int* neighborX,
                                            unsigned int* neighborY,
                                            unsigned int* neighborZ,
                                            unsigned long long numberOfLBnodes,
                                            bool isEvenTimestep);

__global__ void CalcCP27(real* DD,
                                    int* cpIndex,
                                    int nonCp,
                                    double *cpPress,
                                    unsigned int* neighborX,
                                    unsigned int* neighborY,
                                    unsigned int* neighborZ,
                                    unsigned long long numberOfLBnodes,
                                    bool isEvenTimestep);

__global__ void getSendFsPre27(real* DD,
                                          real* bufferFs,
                                          int* sendIndex,
                                          int buffmax,
                                          unsigned int* neighborX,
                                          unsigned int* neighborY,
                                          unsigned int* neighborZ,
                                          unsigned long long numberOfLBnodes,
                                          bool isEvenTimestep);

__global__ void getSendFsPost27(real* DD,
                                           real* bufferFs,
                                           int* sendIndex,
                                           int buffmax,
                                           unsigned int* neighborX,
                                           unsigned int* neighborY,
                                           unsigned int* neighborZ,
                                           unsigned long long numberOfLBnodes,
                                           bool isEvenTimestep);

__global__ void setRecvFsPre27(real* DD,
                                          real* bufferFs,
                                          int* recvIndex,
                                          int buffmax,
                                          unsigned int* neighborX,
                                          unsigned int* neighborY,
                                          unsigned int* neighborZ,
                                          unsigned long long numberOfLBnodes,
                                          bool isEvenTimestep);

__global__ void setRecvFsPost27(real* DD,
                                           real* bufferFs,
                                           int* recvIndex,
                                           int buffmax,
                                           unsigned int* neighborX,
                                           unsigned int* neighborY,
                                           unsigned int* neighborZ,
                                           unsigned long long numberOfLBnodes,
                                           bool isEvenTimestep);

__global__ void getSendGsF3(
    real* G6,
    real* bufferGs,
    int* sendIndex,
    int buffmax,
    unsigned int* neighborX,
    unsigned int* neighborY,
    unsigned int* neighborZ,
    unsigned long long numberOfLBnodes,
    bool isEvenTimestep);

__global__ void setRecvGsF3(
    real* G6,
    real* bufferGs,
    int* recvIndex,
    int buffmax,
    unsigned int* neighborX,
    unsigned int* neighborY,
    unsigned int* neighborZ,
    unsigned long long numberOfLBnodes,
    bool isEvenTimestep);

__global__ void WallFunction27(     real* vx,
                                            real* vy,
                                            real* vz,
                                            real* DD,
                                            int* k_Q,
                                            real* QQ,
                                            unsigned int numberOfBCnodes,
                                            real om1,
                                            unsigned int* neighborX,
                                            unsigned int* neighborY,
                                            unsigned int* neighborZ,
                                            unsigned long long numberOfLBnodes,
                                            bool isEvenTimestep);

__global__ void LBSetOutputWallVelocitySP27( real* vxD,
                                                        real* vyD,
                                                        real* vzD,
                                                        real* vxWall,
                                                        real* vyWall,
                                                        real* vzWall,
                                                        int numberOfWallNodes,
                                                        int* kWallNodes,
                                                        real* rhoD,
                                                        real* pressD,
                                                        unsigned int* geoD,
                                                        unsigned int* neighborX,
                                                        unsigned int* neighborY,
                                                        unsigned int* neighborZ,
                                                        unsigned long long numberOfLBnodes,
                                                        real* DD,
                                                        bool isEvenTimestep);

__global__ void GetVeloforForcing27( real* DD,
                                                int* bcIndex,
                                                int nonAtBC,
                                                real* Vx,
                                                real* Vy,
                                                real* Vz,
                                                unsigned int* neighborX,
                                                unsigned int* neighborY,
                                                unsigned int* neighborZ,
                                                unsigned long long numberOfLBnodes,
                                                bool isEvenTimestep);

__global__ void InitParticles( real* coordX,
                                          real* coordY,
                                          real* coordZ,
                                          real* coordParticleXlocal,
                                          real* coordParticleYlocal,
                                          real* coordParticleZlocal,
                                          real* coordParticleXglobal,
                                          real* coordParticleYglobal,
                                          real* coordParticleZglobal,
                                          real* veloParticleX,
                                          real* veloParticleY,
                                          real* veloParticleZ,
                                          real* randArray,
                                          unsigned int* particleID,
                                          unsigned int* cellBaseID,
                                          unsigned int* bcMatD,
                                          unsigned int* neighborX,
                                          unsigned int* neighborY,
                                          unsigned int* neighborZ,
                                          unsigned int* neighborWSB,
                                          int level,
                                          unsigned int numberOfParticles,
                                          unsigned long long numberOfLBnodes);

__global__ void MoveParticles( real* coordX,
                                          real* coordY,
                                          real* coordZ,
                                          real* coordParticleXlocal,
                                          real* coordParticleYlocal,
                                          real* coordParticleZlocal,
                                          real* coordParticleXglobal,
                                          real* coordParticleYglobal,
                                          real* coordParticleZglobal,
                                          real* veloParticleX,
                                          real* veloParticleY,
                                          real* veloParticleZ,
                                          real* DD,
                                          real  omega,
                                          unsigned int* particleID,
                                          unsigned int* cellBaseID,
                                          unsigned int* bcMatD,
                                          unsigned int* neighborX,
                                          unsigned int* neighborY,
                                          unsigned int* neighborZ,
                                          unsigned int* neighborWSB,
                                          int level,
                                          unsigned int timestep,
                                          unsigned int numberOfTimesteps,
                                          unsigned int numberOfParticles,
                                          unsigned long long numberOfLBnodes,
                                          bool isEvenTimestep);

__global__ void MoveParticlesWithoutBCs(   real* coordX,
                                                      real* coordY,
                                                      real* coordZ,
                                                      real* coordParticleXlocal,
                                                      real* coordParticleYlocal,
                                                      real* coordParticleZlocal,
                                                      real* coordParticleXglobal,
                                                      real* coordParticleYglobal,
                                                      real* coordParticleZglobal,
                                                      real* veloParticleX,
                                                      real* veloParticleY,
                                                      real* veloParticleZ,
                                                      real* DD,
                                                      real  omega,
                                                      unsigned int* particleID,
                                                      unsigned int* cellBaseID,
                                                      unsigned int* bcMatD,
                                                      unsigned int* neighborX,
                                                      unsigned int* neighborY,
                                                      unsigned int* neighborZ,
                                                      unsigned int* neighborWSB,
                                                      int level,
                                                      unsigned int timestep,
                                                      unsigned int numberOfTimesteps,
                                                      unsigned int numberOfParticles,
                                                      unsigned long long numberOfLBnodes,
                                                      bool isEvenTimestep);

__global__ void initRandom(curandState* state);

__global__ void generateRandomValues(curandState* state,
                                                real* randArray);

__global__ void CalcTurbulenceIntensity(
   real* vxx,
   real* vyy,
   real* vzz,
   real* vxy,
   real* vxz,
   real* vyz,
   real* vx_mean,
   real* vy_mean,
   real* vz_mean,
   real* DD,
   uint *typeOfGridNode,
   unsigned int* neighborX,
   unsigned int* neighborY,
   unsigned int* neighborZ,
   unsigned long long numberOfLBnodes,
   bool isEvenTimestep);

#endif
